/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the QtSql module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
**
**
**
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "qsqlrelationaltablemodelmod1.h"

#include "qhash.h"
#include "qstringlist.h"
#include "qsqldatabase.h"
#include "qsqldriver.h"
#include "qsqlerror.h"
#include "qsqlfield.h"
#include "qsqlindex.h"
#include "qsqlquery.h"
#include "qsqlrecord.h"

#include "qsqltablemodel_p.h"
#include "qsqlrelationaltablemodelmod1_p.h"

#include "qdebug.h"

QT_BEGIN_NAMESPACE


class QRelatedTableModelMod1;

class QRelatedTableModelMod1 : public QSqlTableModel
{
public:
    QRelatedTableModelMod1(QRelationMod1 *rel, QObject *parent = 0, QSqlDatabase db = QSqlDatabase());
    bool select();
private:
    bool firstSelect;
    QRelationMod1 *relation;
};



/*
    A QRelation must be initialized before it is considered valid.
    Note: population of the model and dictionary are kept separate
          from initialization, and are populated on an as needed basis.
*/
void QRelationMod1::init(QSqlRelationalTableModelMod1 *parent,
                         const QSqlRelationMod1 &relation,
                         int &addIdColumn)
{
    Q_ASSERT(parent != NULL);
    m_parent = parent;
    rel = relation;
    id_column = addIdColumn;
}

bool QRelationMod1::isValid()
{
    return (rel.isValid() && m_parent != NULL);
}


ExColumns::ExColumns(RELATION_TYPE rel_Type,
                     const QString &tableFrom,
                     const QString &fieldFrom,
                     const QString &funcName,
                     const QString foreignKeyField,
                     const QString fieldNameTo,
                     int srcCol):
    sTableFrom(tableFrom),
    sFieldFrom(fieldFrom),
    sFuncName(funcName),
    sForeignKeyField(foreignKeyField),
    sFieldNameTo(fieldNameTo),
    cFormat(0),
    precision(0),
    alignment(Qt::AlignRight)
{
    relType = rel_Type;
    origCol = srcCol;
}

/*QAddColumns::QAddColumns():
    cFormat(0),
    precision(0)
{
    qDebug() << "QAddColumns::QAddColumns";
}*/

QRelatedTableModelMod1::QRelatedTableModelMod1(QRelationMod1 *rel, QObject *parent, QSqlDatabase db) :
    QSqlTableModel(parent, db), firstSelect(true), relation(rel)
{
}

bool QRelatedTableModelMod1::select()
{
    /*if (firstSelect) {
        firstSelect = false;
        return QSqlTableModel::select();
    }
    relation->clearDictionary();
    bool res = QSqlTableModel::select();
    if (res)
        relation->populateDictionary();
        */
    return true;//res;
}


void QSqlRelationalTableModelMod1Private::clearChanges()
{
    for (int i = 0; i < relations.count(); ++i)
    {
        QRelationMod1 &rel = relations[i];
        //rel.clear();
    }
}

void QSqlRelationalTableModelMod1Private::revertCachedRow(int row)
{
    QSqlTableModelPrivate::revertCachedRow(row);
}

int QSqlRelationalTableModelMod1Private::nameToIndex(const QString &name) const
{
    QString fieldname = name;

    if (db.driver()->isIdentifierEscaped(fieldname, QSqlDriver::FieldName))
        fieldname = db.driver()->stripDelimiters(fieldname, QSqlDriver::FieldName);

    return baseRec.indexOf(fieldname);
}

void QSqlRelationalTableModelMod1Private::clearEditBuffer()
{
    editBuffer = baseRec; // do not delete, this inits the editBuffer
    clearGenerated(editBuffer);
    //editBufferEx =
}


QSqlRelationalTableModelMod1::QSqlRelationalTableModelMod1(QObject *parent, QSqlDatabase db)
    : QSqlTableModel(*new QSqlRelationalTableModelMod1Private, parent, db)
{
}

QSqlRelationalTableModelMod1::QSqlRelationalTableModelMod1(QSqlRelationalTableModelMod1Private &dd,
                                                           QObject *parent,
                                                           QSqlDatabase db)
    : QSqlTableModel(dd,
                     parent,
                     db)
{

}


QSqlRelationalTableModelMod1::~QSqlRelationalTableModelMod1()
{
}


QVariant QSqlRelationalTableModelMod1::data(const QModelIndex &idx, int role) const
{
    Q_D(const QSqlRelationalTableModelMod1);

    int row = idx.row();

    int col = idx.column();
    int max = d->baseRec.count();

    if(col >= max )
    {
        int exCol = col - max;

        if( exCol < d->ex_columns.count())
        {
            if(role == Qt::TextAlignmentRole)
            {
                if(d->ex_columns[exCol].alignment != 0)
                    return d->ex_columns[exCol].alignment;

            }
            else if(role == Qt::DisplayRole)
            {
                char ch = d->ex_columns.at(exCol).cFormat;

                if( ch != 0)
                {
                    QVariant vv = QSqlTableModel::data(idx, role);

                    if(vv.type() == QVariant::Double)
                    {
                        return QString::number(vv.toDouble(),
                                               d->ex_columns.at(exCol).cFormat,
                                               d->ex_columns.at(exCol).precision) ;
                    }
                    return vv;
                }
            }
        }
        return QSqlTableModel::data(idx, role);
    }
    else if(role == Qt::UserRole)
    {
        if(d->relations[col].isValid())
        {
            int id_column = d->relations[col].id_column;

            if ( id_column >=0)
                return data(index(row , id_column));
            else
                return 0;
        }
    }


    return QSqlTableModel::data(idx, role);
}

Qt::ItemFlags QSqlRelationalTableModelMod1::flags(const QModelIndex &index) const
{
    Q_D(const QSqlRelationalTableModelMod1);

    int col = index.column();
    int max = d->baseRec.count();

    if(col >= max)
    {
        int add = col - max;

        if(add < d->ex_columns.count())
        {
            return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
        }
    }
    else
        return QSqlTableModel::flags(index);
}

QSqlRecord QSqlRelationalTableModelMod1::origRecord()
{
    Q_D(const QSqlRelationalTableModelMod1);
    return d->baseRec;
}

bool QSqlRelationalTableModelMod1::setDoubleFormat(int exCol,
                                                   const char &format,
                                                   const int &precision)
{
    Q_D(const QSqlRelationalTableModelMod1);

    if(exCol >=0 && exCol < d->ex_columns.count())
    {

        d->ex_columns[exCol].setFormat(format , precision);
        return true;
    }
    return false;

}

bool QSqlRelationalTableModelMod1::setData(const QModelIndex &idx,
                                           const QVariant &value,
                                           int role)
{
    return QSqlTableModel::setData(idx, value, role);
}


bool QSqlRelationalTableModelMod1::setRelation(int column,
                                               const QSqlRelationMod1 &relation)
{
    Q_D(QSqlRelationalTableModelMod1);

    if (column < 0)
        return false;

    if (d->relations.size() <= column)
        d->relations.resize(column + 1);

    int exCount = d->ex_columns.count();
    int origCount = d->baseRec.count();

    foreach(ExColumns exCol, d->ex_columns)
    {
        if(exCol.origCol == column)
            return false;
    }

    d->ex_columns.append(ExColumns(ExColumns::RELATION_EXT_TEXT,
                                   relation.tableFrom(),
                                   relation.indexFieldFrom(),
                                   relation.displayFieldFrom(),
                                   QString(),
                                   QString(),
                                   column
                                   ));

    int relIdCol = origCount + exCount;

    d->relations[column].init( this , relation , relIdCol);

    return true;
}

bool QSqlRelationalTableModelMod1::setCalcField(const QString &tableFrom,
                                                    const QString &fieldFrom,
                                                    const QString &funcName,
                                                    const QString foreignKeyField,
                                                    const QString fieldNameTo
                                                    )
{
    Q_D(QSqlRelationalTableModelMod1);

    if(tableFrom.isEmpty() || tableFrom.isNull())
        return false;
    if(fieldFrom.isEmpty() || fieldFrom.isNull())
        return false;
    if(funcName.isEmpty() || funcName.isNull())
        return false;

    d->ex_columns.append(ExColumns(
                             ExColumns::RELATION_EXT_CALCULATION,
                             tableFrom,
                             fieldFrom,
                             funcName,
                             foreignKeyField,
                             fieldNameTo));


    return true;
}

QSqlRelationMod1 QSqlRelationalTableModelMod1::relation(int column) const
{
    Q_D(const QSqlRelationalTableModelMod1);
    return d->relations.value(column).rel;
}

QString QSqlRelationalTableModelMod1Private::relationField(const QString &tableName,
                                                           const QString &fieldName) const
{
    QString ret;

    ret.reserve(tableName.size() + fieldName.size() + 1);
    ret.append(tableName).append(QLatin1Char('.')).append(fieldName);

    return ret;
}


QSqlTableModel *QSqlRelationalTableModelMod1::relationModel(int column) const
{
    Q_D(const QSqlRelationalTableModelMod1);

    if ( column < 0 || column >= d->relations.count())
        return 0;

    QRelationMod1 &relation = const_cast<QSqlRelationalTableModelMod1Private *>(d)->relations[column];

    if (!relation.isValid())
        return 0;

    return relation.model;
}


void QSqlRelationalTableModelMod1::revertRow(int row)
{
    QSqlTableModel::revertRow(row);
}


bool QSqlRelationalTableModelMod1::select()
{
    Q_D(QSqlRelationalTableModelMod1);

    QString query = selectStatement();

    if (query.isEmpty())
        return false;

    revertAll();

    QSqlQuery qu(query, d->db);
    setQuery(qu);

    if (!qu.isActive() || lastError().isValid())
    {
        // something went wrong - revert to non-select state
        d->initRecordAndPrimaryIndex();
        return false;
    }

    int origColumnCount = origRecord().count();

    for(uint exCol=0; exCol < exColCount(); exCol++)
    {

        QString header = exColName(exCol);

        if( header != QString())
        {
            bool b =setHeaderData( origColumnCount + exCol, Qt::Horizontal, header);
        }
    }

    return true;
}


void QSqlRelationalTableModelMod1::setTable(const QString &table)
{
    Q_D(QSqlRelationalTableModelMod1);

    // memorize the table before applying the relations
    d->baseRec = d->db.record(table);
    d->ex_columns.clear();

    //qDebug()  <<  "QSqlRelationalTableModelMod1::setTable " << table;

    QSqlTableModel::setTable(table);
}

void QSqlRelationalTableModelMod1Private::translateFieldNames(int row, QSqlRecord &values) const
{
    Q_Q(const QSqlRelationalTableModelMod1);

    for (int i = 0; i < values.count(); ++i)
    {
        int realCol = q->indexInQuery(q->createIndex(row, i)).column();

        if (realCol != -1 && relations.value(realCol).isValid())
        {
            QVariant v = values.value(i);
            bool gen = values.isGenerated(i);
            values.replace(i, baseRec.field(realCol));
            values.setValue(i, v);
            values.setGenerated(i, gen);
        }
    }
}


bool QSqlRelationalTableModelMod1::updateRowInTable(int row, const QSqlRecord &values)
{
    qDebug() << "QSqlRelationalTableModelMod1::updateRowInTable values " << values;

    Q_D(QSqlRelationalTableModelMod1);

    QSqlRecord rec = values;
    d->translateFieldNames(row, rec);

    return QSqlTableModel::updateRowInTable(row, rec);
}

bool QSqlRelationalTableModelMod1::insertRowIntoTable(const QSqlRecord &values)
{
    Q_D(QSqlRelationalTableModelMod1);

    qDebug() << "QSqlRelationalTableModelMod1::insertRowIntoTable values " << values;

    QSqlRecord rec = values;
    d->translateFieldNames(0, rec);

    qDebug() << "QSqlRelationalTableModelMod1::insertRowIntoTable values " << values;

    return QSqlTableModel::insertRowIntoTable(rec);
}


QString QSqlRelationalTableModelMod1::orderByClause() const
{
    Q_D(const QSqlRelationalTableModelMod1);

    const QSqlRelationMod1 rel = d->relations.value(d->sortColumn).rel;
    if (!rel.isValid())
        return QSqlTableModel::orderByClause();

    QString s = QLatin1String("ORDER BY ");
    s.append(d->relationField(QLatin1String("relTblAl_") + QString::number(d->sortColumn),
                              rel.displayFieldFrom()));
    s += d->sortOrder == Qt::AscendingOrder ? QLatin1String(" ASC") : QLatin1String(" DESC");
    return s;
}


bool QSqlRelationalTableModelMod1::removeColumns(int column, int count, const QModelIndex &parent)
{
    Q_D(QSqlRelationalTableModelMod1);

    if (parent.isValid() || column < 0 || column + count > d->rec.count())
        return false;

    for (int i = 0; i < count; ++i)
    {
        d->baseRec.remove(column);
        if (d->relations.count() > column)
            d->relations.remove(column);
    }
    return QSqlTableModel::removeColumns(column, count, parent);
}



QString QSqlRelationalTableModelMod1::selectStatement() const
{
    Q_D(const QSqlRelationalTableModelMod1);
    QString query;

    if (tableName().isEmpty())
        return query;

    /*if (d->relations.isEmpty())
        return QSqlTableModel::selectStatement();*/

    QString sTablesList;
    QString sFieldsList;
    QString where;

    // This is impotant d->baseRec ! because of d->rec maybe contains modified field names already after first select
    QSqlRecord origRec = d->baseRec; // this is after setTable created

    QStringList lstTables;
    const QRelationMod1 nullRelation;

    // Count how many times each field name occurs in the record
    QHash<QString, int> fieldNames;
    QStringList fieldList;

    bool bFieldsFullName = false;

    if( d->ex_columns.count()>0)
        bFieldsFullName = true;

    for (int i = 0; i < origRec.count(); ++i)
    {
        QSqlRelationMod1 relation = d->relations.value(i, nullRelation).rel;

        QString name;

        if (relation.isValid())
        {
            // Count the display column name, not the original foreign key
            name = relation.displayFieldFrom();

            if (d->db.driver()->isIdentifierEscaped(name, QSqlDriver::FieldName))
                name = d->db.driver()->stripDelimiters(name, QSqlDriver::FieldName);

            QSqlRecord recOrig = database().record(tableName());

            for (int i = 0; i < recOrig.count(); ++i)
            {
                // test field name exists
                if (name.compare(recOrig.fieldName(i), Qt::CaseInsensitive) == 0)
                {
                    name.prepend(QLatin1String("_")).prepend(relation.tableFrom());
                    break;
                }
            }
        }
        else
            name = origRec.fieldName(i);

        if( bFieldsFullName )
            name.prepend(QLatin1String(".")).prepend(tableName());


        fieldNames.insert(name, fieldNames.value(name, 0) + 1);
        fieldList.append(name);
    }
    //qDebug() << "\nfieldNames " << fieldNames;
    //qDebug() << "\nfieldList " << fieldList;
    //qDebug() << "\nsFieldsList " << sFieldsList;


    // --------------------------------------------------
    //              ORIGINAL FILEDS
    // --------------------------------------------------

    for (int col = 0; col < origRec.count(); ++col)
    {
        QSqlRelationMod1 relation = d->relations.value(col, nullRelation).rel;

        if ( ! sFieldsList.isEmpty())
            sFieldsList.append(QLatin1String(", \n"));

        if (relation.isValid())
        {
            QString relTableAlias = QString::fromLatin1("relTblAl_%1").arg(col);

            sFieldsList.append(d->relationField(relTableAlias , relation.displayFieldFrom()));

            sFieldsList.append(QString::fromLatin1(" AS %1_%2_%3").
                               arg(tableName()).
                               arg(origRec.fieldName(col)).
                               arg(col));

        }
        else
        {
            sFieldsList.append(tableName()).
                    append(QLatin1String(".")).
                    append(origRec.fieldName(col));
        }
    }

    //qDebug() << "\n2 fieldNames " << fieldNames;
    //qDebug() << "\n2 fieldList " << fieldList;
    //qDebug() << "\n2 sFieldsList " << sFieldsList;

    QString sAddFields;

    //-------------------------------------------------------
    //                      EXT TEXT / CALC
    //-------------------------------------------------------

    QString sExtFieldNameAdds;

    for(int col=0; col < d->ex_columns.count(); col++)
    {
        ExColumns exCol = d->ex_columns.at(col);

        if(exCol.relType == ExColumns::RELATION_EXT_TEXT)
        {
            QString relTableAlias = QString::fromLatin1("relTblAl_%1").arg(exCol.origCol);


            sAddFields.append(QLatin1String(", \n")).
                    append(relTableAlias).
                    append(QLatin1String(".")).
                    append(exCol.sFieldFrom).
                    append(QLatin1String(" as ")).
                    append(exCol.sTableFrom).
                    append(QLatin1String("_")).
                    append(exCol.sFieldFrom).
                    append(QLatin1String("_"));


        }
        else if(exCol.relType == ExColumns::RELATION_EXT_CALCULATION)
        {
            QStringList lst = exCol.sFieldFrom.split(QLatin1String(","));

            QString fields;

            sExtFieldNameAdds.
                    append(QLatin1String(", \n")).
                    append(exCol.sFuncName).
                    append(QLatin1String("("));

            for(int ii=0; ii < lst.count(); ii++) // this is a sum by some fields
            {

                sExtFieldNameAdds.
                        append(exCol.sTableFrom).
                        append(QLatin1String(".")).
                        append(lst.at(ii));

                if(exCol.sFieldNameTo.isEmpty())
                    fields.append(lst.at(ii));

                if(ii < lst.count() -1)
                {
                    sExtFieldNameAdds.append(QLatin1String(" + "));

                    if(exCol.sFieldNameTo.isEmpty())
                        fields.append(QLatin1String("_"));
                }
            }
            sExtFieldNameAdds.append(QLatin1String(") as "));

            if(exCol.sFieldNameTo.isNull())
                sExtFieldNameAdds.append(exCol.sTableFrom).append(QLatin1String("_")).append(fields);
            else
                sExtFieldNameAdds.append(exCol.sFieldNameTo);
        }
    }

    QString sExtLeftJoins;
    QStringList joinedTables;

    //-------------------------------------------------------
    //                      JOIN
    //-------------------------------------------------------


    for(int ii=0; ii < d->ex_columns.count(); ii++)
    {
        ExColumns addCol = d->ex_columns.at(ii);

        if( addCol.relType == ExColumns::RELATION_EXT_TEXT)
        {

            if( ! joinedTables.contains(addCol.sTableFrom))
            {
                int fggf= addCol.origCol;

                QString relTableAlias = QString::fromLatin1("relTblAl_%1").arg(addCol.origCol);

                sExtLeftJoins.append(QLatin1String(" \nLEFT JOIN ")).
                        append(addCol.sTableFrom).
                        append(QLatin1String(" ")).
                        append(relTableAlias).

                        append(QLatin1String(" ON ")).

                        append(tableName()).
                        append(QLatin1String(".")).
                        append(origRec.fieldName(addCol.origCol)).

                        append(QLatin1String("=")).

                        append(relTableAlias).
                        append(QLatin1String(".id"));

                joinedTables.append(addCol.sTableFrom);
            }
        }
        else  if( addCol.relType == ExColumns::RELATION_EXT_CALCULATION)
        {
            if( ! joinedTables.contains(addCol.sTableFrom))
            {
                sExtLeftJoins.append(QLatin1String(" \nLEFT JOIN ")).
                        append(addCol.sTableFrom).
                        append(QLatin1String(" ")).

                        append(QLatin1String(" ON ")).

                        append(tableName()).
                        append(QLatin1String(".")).
                        append(QLatin1String("id")).

                        append(QLatin1String("=")).

                        append(addCol.sTableFrom).
                        append(QLatin1String(".")).
                        append(addCol.sForeignKeyField);

                joinedTables.append(addCol.sTableFrom);
            }
        }
    }

    //------------------------------------------------------------

    sTablesList.append(lstTables.join(QLatin1String(" ")));

    if (sFieldsList.isEmpty())
        return query;

    sTablesList.prepend(tableName());
    query.append(QLatin1String("SELECT "));

    if( ! sExtFieldNameAdds.isEmpty())
    {
        sFieldsList.append(sExtFieldNameAdds);
    }


    query.append(sFieldsList).append(sAddFields).append(QLatin1String(" \nFROM ")).append(sTablesList);


    if( ! sExtLeftJoins.isEmpty())
    {
        query.append(sExtLeftJoins);
    }

    if( ! sExtFieldNameAdds.isEmpty())
    {
        query.append(QLatin1String(" \nGROUP BY ")).append(tableName()).append(QLatin1String(".id"));
    }

    // --------------------- WHERE ------------------------

    if (!filter().isEmpty() )
    {

        query.append(QLatin1String(" \nWHERE ("));
        query.append(filter());
        query.append(QLatin1String(")"));
    }

    // ---------------------- ORDER -------------------------

    QString orderBy = orderByClause();

    if (!orderBy.isEmpty())
        query.append(QLatin1Char(' ')).append(orderBy);

    return query;
}

void QSqlRelationalTableModelMod1::setAlignment(int exCol , Qt::Alignment align)
{
    Q_D(QSqlRelationalTableModelMod1);

    if(exCol < d->ex_columns.count() )
        d->ex_columns[exCol].setAlignment(align);
}

ExColumns::RELATION_TYPE QSqlRelationalTableModelMod1::exColType(int exCol)
{
    Q_D(QSqlRelationalTableModelMod1);

    //qDebug() << " d->addColumns " << d->addColumns;

    if(exCol < d->ex_columns.count())
    {
        ExColumns::RELATION_TYPE tt = d->ex_columns[exCol].relType;
        return tt;
    }
    return ExColumns::RELATION_EXT_UNKNOUWN;
}

QString  QSqlRelationalTableModelMod1::exColName(int exCol)
{
    Q_D(QSqlRelationalTableModelMod1);

    if(exCol < d->ex_columns.count())
    {
        QString str = d->ex_columns[exCol].sFieldNameTo;
        return str;
    }
    return QString();
}

QString  QSqlRelationalTableModelMod1::exStr(int exCol)
{
    Q_D(QSqlRelationalTableModelMod1);

    if(exCol < d->ex_columns.count())
    {
        QString str;

        str.append(d->ex_columns[exCol].sTableFrom).
                append(QLatin1Char(' ')).
                append(d->ex_columns[exCol].sFieldFrom).
                append(QLatin1Char(' ')).append(d->ex_columns[exCol].sFuncName);

        return str;
    }
}

int QSqlRelationalTableModelMod1::exColCount()
{
    Q_D(QSqlRelationalTableModelMod1);

    return d->ex_columns.count();
}


int QSqlRelationalTableModelMod1::relColumnId(int relCol)
{
    Q_D(QSqlRelationalTableModelMod1);

    QSqlRelationMod1 relation = d->relations.value(relCol).rel;

    if(relation.isValid())
        return d->relations[relCol].id_column;
    else
        return 0;
}


QT_END_NAMESPACE

