/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the QtSql module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** GNU Lesser General Public License Usage
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this
** file. Please review the following information to ensure the GNU Lesser
** General Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU General
** Public License version 3.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of this
** file. Please review the following information to ensure the GNU General
** Public License version 3.0 requirements will be met:
** http://www.gnu.org/copyleft/gpl.html.
**
** Other Usage
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
**
**
**
**
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef QSQLRELATIONALTABLEMODELMOD1_H
#define QSQLRELATIONALTABLEMODELMOD1_H

#include <QtSql/qsqltablemodel.h>

QT_BEGIN_HEADER

QT_BEGIN_NAMESPACE

QT_MODULE(Sql)


class QSqlRelationalTableModelMod1Private;

class Q_SQL_EXPORT QSqlRelationMod1
{
public:
    enum JoinMode {
        InnerJoin,
        LeftJoin
    };
public:
    QSqlRelationMod1() {}

    QSqlRelationMod1(const QString &sTableFrom,
                     const QString &sIndexFieldFrom,
                     const QString &sDisplayFieldFrom,
                     JoinMode join_mode = QSqlRelationMod1::LeftJoin)
        : sTableFrom(sTableFrom),
          sIndexFieldFrom(sIndexFieldFrom),
          sDisplayFieldFrom(sDisplayFieldFrom)
    {
        joinMode = join_mode;
    }

    inline QString tableFrom() const
    { return sTableFrom; }

    inline QString indexFieldFrom() const
    { return sIndexFieldFrom; }

    inline QString displayFieldFrom() const
    { return sDisplayFieldFrom; }

    inline bool isValid() const
    { return !(sTableFrom.isEmpty() || sIndexFieldFrom.isEmpty() || sDisplayFieldFrom.isEmpty()); }

private:
    JoinMode joinMode;
    QString sTableFrom, sIndexFieldFrom, sDisplayFieldFrom;
};

class QSqlRelationalTableModelMod1;
class QRelatedTableModelMod1;

struct QRelationMod1
{

public:
    QRelationMod1():
        model(0),
        m_parent(0),
        m_dictInitialized(false),
        id_column(-1)
    {}

    void init(QSqlRelationalTableModelMod1 *parent,
              const QSqlRelationMod1 &relation,
              int &addIdColumn);

    bool isValid();

    QSqlRelationMod1 rel;
    QRelatedTableModelMod1 *model;
    //QHash<QString, QVariant> dictionary;//maps keys to display values

    int id_column;

private:
    QSqlRelationalTableModelMod1 *m_parent;
    bool m_dictInitialized;
    //COLUMN_TYPE *relType;
};

struct ExColumns
{

public:
    enum RELATION_TYPE
    {
        RELATION_EXT_UNKNOUWN=0,
        RELATION_EXT_TEXT=1,
        RELATION_EXT_CALCULATION=2
    };

    //QAddColumns();
    ExColumns(
            RELATION_TYPE relType,
            const QString &tableFrom,
            const QString &fieldFrom,
            const QString &funcName,
            const QString foreignKeyField=QString(),
            const QString fieldNameTo=QString(),
            int origCol = -1);

    QRelatedTableModelMod1 *model;
    int origCol;
    RELATION_TYPE relType;
    QString sTableFrom;
    QString sForeignKeyField;
    QString sFieldFrom;
    QString sFuncName;
    QString sFieldNameTo;

    char cFormat;
    int precision;

    void setFormat(const char &format, const int &precision_)
    {
        cFormat = format;
        precision = precision_;
    }

    void setAlignment(Qt::Alignment align = Qt::AlignLeft)
    {
        alignment = align;
    }
    Qt::Alignment alignment;

private:
    QSqlRelationalTableModelMod1 *m_parent;

};


typedef struct CALC_COLUMN
{
    const QString     sFromTable;
    const QString     sForeignKeyField;
    const QString     sFieldNameFrom;
    const QString     sCalcFunc;
    const QString     sFieldNameTo;
};

class Q_SQL_EXPORT QSqlRelationalTableModelMod1: public QSqlTableModel
{
    Q_OBJECT

public:

    explicit QSqlRelationalTableModelMod1(QObject *parent = 0,
                                          QSqlDatabase db = QSqlDatabase());

    virtual ~QSqlRelationalTableModelMod1();

    //QList <CALC_COLUMN> calcFieldsFromExtTableExists;

    bool setCalcField(const QString &tableFrom,
                          const QString &fieldFrom,
                          const QString &funcName,
                          const QString foreignKeyField,
                          const QString fieldNameTo=QString());


    virtual QString selectStatement() const;

    virtual QVariant data(const QModelIndex &item, int role = Qt::DisplayRole) const; // !! add virtual
    virtual bool setData(const QModelIndex &item, const QVariant &value, int role = Qt::EditRole); // !! add virtual
    virtual Qt::ItemFlags flags(const QModelIndex &index) const;

    bool removeColumns(int column, int count, const QModelIndex &parent = QModelIndex());

    bool setRelation(int column, const QSqlRelationMod1 &relation);

    bool select();


    void setTable(const QString &tableName);

    QSqlRelationMod1 relation(int column) const;

    virtual QSqlTableModel *relationModel(int column) const;

    //void setJoinMode( QSqlRelationalTableModelMod1::JoinMode joinMode );

    bool setDoubleFormat(int col,
                         const char &format,
                         const int &precision);

    void setAlignment(int exCol , Qt::Alignment align);

    QSqlRecord origRecord();

    ExColumns::RELATION_TYPE exColType(int exCol);
    int exColCount();
    QString exStr(int exCol);
    QString  exColName(int exCol);

    int relColumnId(int relCol);


public Q_SLOTS:
    void revertRow(int row);

protected:

    //!! ------------- my addition -----------------------
    QSqlRelationalTableModelMod1(QSqlRelationalTableModelMod1Private &dd,
                                 QObject *parent = 0,
                                 QSqlDatabase db = QSqlDatabase());
    //------------------------------------------------------------

    bool updateRowInTable(int row, const QSqlRecord &values);
    bool insertRowIntoTable(const QSqlRecord &values);
    QString orderByClause() const;

private:

    //QMap<int,RELATION_TYPE*> mdl_columns;

    Q_DECLARE_PRIVATE(QSqlRelationalTableModelMod1)
};

QT_END_NAMESPACE

QT_END_HEADER

#endif // QSQLRELATIONALTABLEMODELMOD1_H
